use std::{
    collections::HashSet,
    net::SocketAddr,
    sync::{Arc, LazyLock, Weak},
};

use auth::authenticate_socket;
use axum::{
    extract::{
        ws::{Message, WebSocket},
        Path, State, WebSocketUpgrade,
    },
    http::header,
    response::IntoResponse,
    routing::get,
    Router,
};
use futures_util::{stream::SplitSink, SinkExt, StreamExt, TryStreamExt};
use mongodb::{
    bson::{oid::ObjectId, Bson, Document},
    Client, Collection, Database,
};
use parking_lot::RwLock;
use serde::{Deserialize, Serialize};
use serde_json::json;
use uuid::Uuid;
use weak_table::PtrWeakHashSet;

mod auth;

static CONNECTED_MAP: LazyLock<RwLock<HashSet<(MinecraftServer, Uuid)>>> =
    LazyLock::new(|| RwLock::new(HashSet::new()));

static REQWEST_CLIENT: LazyLock<reqwest::Client> = LazyLock::new(|| {
    reqwest::Client::builder()
        .user_agent("Samsung Smart Fridge")
        .build()
        .expect("Failed to initialize Reqwest")
});

static CONNECTIONS: LazyLock<
    RwLock<
        PtrWeakHashSet<Weak<tokio::sync::Mutex<SplitSink<WebSocket, axum::extract::ws::Message>>>>,
    >,
> = LazyLock::new(|| RwLock::new(PtrWeakHashSet::new()));

#[tokio::main]
async fn main() -> eyre::Result<()> {
    let mongo = Client::with_uri_str("mongodb://localhost:27017")
        .await
        .expect("Unable to connect to mongodb database");

    let app = Router::new()
        .route("/api/cape/:uuid", get(cape_lookup_endpoint))
        .route("/api/status/:uuid", get(status_lookup_endpoint))
        .route("/api/cape_status/:uuid", get(cape_status_lookup_endpoint))
        .route("/api/ws", get(socket_endpoint));
    let listener = tokio::net::TcpListener::bind("127.0.0.1:8585")
        .await
        .unwrap();
    axum::serve(listener, app.with_state(mongo.database("sun-server")))
        .await
        .unwrap();

    Ok(())
}

#[derive(Serialize, Deserialize, Hash, Eq, PartialEq, Clone)]
#[serde(tag = "type", content = "value", rename_all = "snake_case")]
enum MinecraftServer {
    NetworkServer(String),
    SocketAddr(SocketAddr),
}

#[derive(Serialize, Deserialize)]
struct Cape {
    _id: ObjectId,
    uuid: String,
    cape: String,
}

#[derive(Serialize, Deserialize)]
struct User {
    _id: ObjectId,
    uuid: String,
    status: Status,
}

#[derive(Serialize, Deserialize)]
#[serde(rename_all = "snake_case")]
enum Status {
    Connected,
    ConnectedSpecial,
    ConnectedAdmin,
    Disconnected,
    NonSun,
}

async fn cape_lookup(uuid: &Uuid, mongo: &Database) -> Option<Cape> {
    let capes: Collection<Cape> = mongo.collection("capes");
    capes
        .find_one(Document::from_iter([(
            "uuid".into(),
            Bson::String(uuid.to_string()),
        )]))
        .await
        .unwrap()
}

async fn status_lookup(server: &MinecraftServer, uuid: &Uuid, mongo: &Database) -> Option<Status> {
    let users: Collection<User> = mongo.collection("users");
    let user = users
        .find_one(Document::from_iter([(
            "uuid".into(),
            Bson::String(uuid.to_string()),
        )]))
        .await
        .unwrap();
    let Some(user) = user else {
        return None;
    };
    let is_connected = CONNECTED_MAP
        .read()
        .contains(&(server.clone(), uuid.to_owned()));
    Some(if is_connected {
        user.status
    } else {
        Status::Disconnected
    })
}

async fn cape_lookup_endpoint(
    Path(uuid): Path<Uuid>,
    State(mongo): State<Database>,
) -> impl IntoResponse {
    let cape_str = cape_lookup(&uuid, &mongo)
        .await
        .map(|c| c.cape)
        .unwrap_or(String::new());
    (
        [(header::CONTENT_TYPE, "application/json")],
        json!({"cape": cape_str}).to_string(),
    )
}

async fn status_lookup_endpoint(
    Path((server, uuid)): Path<(MinecraftServer, Uuid)>,
    State(mongo): State<Database>,
) -> impl IntoResponse {
    let status = status_lookup(&server, &uuid, &mongo)
        .await
        .unwrap_or(Status::NonSun);
    (
        [(header::CONTENT_TYPE, "application/json")],
        json!({"status": status}).to_string(),
    )
}

async fn cape_status_lookup_endpoint(
    Path((server, uuid)): Path<(MinecraftServer, Uuid)>,
    State(mongo): State<Database>,
) -> impl IntoResponse {
    let cape_str = cape_lookup(&uuid, &mongo)
        .await
        .map(|c| c.cape)
        .unwrap_or(String::new());
    let status = status_lookup(&server, &uuid, &mongo)
        .await
        .unwrap_or(Status::NonSun);
    (
        [(header::CONTENT_TYPE, "application/json")],
        json!({"cape": cape_str, "status": status}).to_string(),
    )
}

#[axum::debug_handler]
async fn socket_endpoint(ws: WebSocketUpgrade) -> impl IntoResponse {
    ws.on_upgrade(|socket| handle_socket(socket))
}

#[derive(Serialize, Deserialize, Clone)]
#[serde(rename_all = "snake_case")]
enum StatusType {
    Connect,
    Disconnect,
}

#[derive(Serialize, Deserialize, Clone)]
struct StatusPacket {
    server: MinecraftServer,
    status_type: StatusType,
}

#[derive(Serialize, Deserialize, Clone)]
struct ChatPacket {
    from: String,
    contents: String,
}

#[derive(Serialize, Deserialize, Clone)]
#[serde(tag = "op", content = "d", rename_all = "snake_case")]
enum Packet {
    Status(StatusPacket),
    Chat(ChatPacket),
}

async fn handle_socket(mut socket: WebSocket) {
    let player_info = match authenticate_socket(&mut socket).await {
        Ok(player_info) => player_info,
        Err(err) => {
            eprintln!("failed to authenticate socket {err:?}");
            return;
        }
    };

    let (socket_w, mut socket_r) = socket.split();
    let socket_w = Arc::new(tokio::sync::Mutex::new(socket_w));
    CONNECTIONS.write().insert(socket_w.clone().into());

    while let Some(message) = socket_r.try_next().await.ok().flatten() {
        let Ok(message) = message.into_text() else {
            continue;
        };
        let Ok(message) = serde_json::from_str::<Packet>(&message) else {
            continue;
        };
        match message {
            Packet::Status(message) => {
                match message.status_type {
                    StatusType::Connect => CONNECTED_MAP
                        .write()
                        .insert((message.server, player_info.id)),
                    StatusType::Disconnect => CONNECTED_MAP
                        .write()
                        .remove(&(message.server, player_info.id)),
                };
            }
            Packet::Chat(message) => {
                if message.from.to_lowercase() != player_info.name.to_lowercase() {
                    return;
                }
                let Ok(message) = serde_json::to_string(&Packet::Chat(message)) else {
                    return;
                };
                let connections = CONNECTIONS.read().iter().collect::<Vec<_>>();
                for connection in connections {
                    let _ = connection
                        .lock()
                        .await
                        .send(Message::Text(message.clone()))
                        .await;
                }
            }
        };
    }

    let to_remove = CONNECTED_MAP
        .read()
        .iter()
        .filter_map(|c| {
            if c.1 == player_info.id {
                Some(c.to_owned())
            } else {
                None
            }
        })
        .collect::<Vec<_>>();
    let mut connected_map_lock = CONNECTED_MAP.write();
    for r in to_remove {
        connected_map_lock.remove(&r);
    }
    // expensive but I don't care atm
    CONNECTIONS.write().remove_expired();
}
